#include "remote.h"

#define FLASH_BASE 0x100000

typedef char int8_t;
typedef unsigned char uint8_t;
typedef unsigned int uint32_t;

#define REG8(address) *((volatile uint8_t*)(address))

extern void sendbyte(int c);
extern int recvbyte(void);
extern void sendstring(const char* s);

extern const uint32_t crctable[];

static uint8_t buffer[SECTORSIZE];

uint32_t recvquad(void)
{
	return recvbyte() |
			(recvbyte() << 8) |
			(recvbyte() << 16) |
			(recvbyte() << 24);
}

uint32_t CRC32(const uint8_t* data, int size)
{
	uint32_t crc = 0xffffffff;

	for (int i=0; i<size; i++)
	{
		uint8_t o = (crc & 0xff) ^ *data++;
		crc = (crc >> 8) ^ crctable[o];
	}

	return ~crc;
}

static void querynand(void)
{
	/* Set flash to Software ID mode. */

	REG8(FLASH_BASE + 0x5555) = 0xaa;
	REG8(FLASH_BASE + 0x2aaa) = 0x55;
	REG8(FLASH_BASE + 0x5555) = 0x90;

	uint8_t manufacturer = REG8(FLASH_BASE + 0);
	uint8_t device = REG8(FLASH_BASE + 1);

	/* Exit software ID mode. */

	REG8(FLASH_BASE + 0) = 0xf0;

	sendbyte(manufacturer);
	sendbyte(device);
}

static void waitforflash(int address)
{
	for (;;)
	{
		int a = REG8(address) & 0x40;
		int b = REG8(address) & 0x40;
		if (a == b)
			break;
	}
}

static void writenand(void)
{
	int sector = recvbyte();

	for (int i = 0; i < SECTORSIZE; i++)
		buffer[i] = recvbyte();

	uint32_t crc = recvquad();
	if (crc != CRC32(buffer, SECTORSIZE))
	{
		sendbyte(RSP_FAILED);
		return;
	}

	int address = FLASH_BASE + (sector * SECTORSIZE);

	/* Erase sector. */

	REG8(FLASH_BASE + 0x5555) = 0xaa;
	REG8(FLASH_BASE + 0x2aaa) = 0x55;
	REG8(FLASH_BASE + 0x5555) = 0x80;

	REG8(FLASH_BASE + 0x5555) = 0xaa;
	REG8(FLASH_BASE + 0x2aaa) = 0x55;
	REG8(address) = 0x30;

	waitforflash(address);

	/* Program sector. */

	for (int i = 0; i < SECTORSIZE; i++)
	{
		REG8(FLASH_BASE + 0x5555) = 0xaa;
		REG8(FLASH_BASE + 0x2aaa) = 0x55;
		REG8(FLASH_BASE + 0x5555) = 0xa0;

		REG8(address + i) = buffer[i];
		waitforflash(address + i);
	}

	/* Verify sector. */

	for (int i = 0; i < SECTORSIZE; i++)
	{
		if (REG8(address + i) != buffer[i])
		{
			sendbyte(RSP_FAILED);
			return;
		}
	}

	sendbyte(RSP_SUCCESS);
}

void remote(void)
{
	sendstring(BANNER);
	sendbyte('\n');

	for (;;)
	{
		switch (recvbyte())
		{
			case CMD_QUERYNAND:
				querynand();
				break;

			case CMD_WRITENAND:
				writenand();
				break;
		}
	}
}
